/*
 * TextureBmpModern.cpp - Version 1.0 - September 14, 2020.
 *
 * Example program illustrating a simple use
 * of Modern OpenGL to use a single texture map
 *
 * Author: Sam Buss
 *
 * Software accompanying POSSIBLE SECOND EDITION TO the book
 *		3D Computer Graphics: A Mathematical Introduction with OpenGL,
 *		by S. Buss, Cambridge University Press, 2003.
 *
 * Software is "as-is" and carries no warranty.  It may be used without
 *   restriction, but if you modify it, please change the filenames to
 *   prevent confusion between different versions.
 * Bug reports: Sam Buss, sbuss@ucsd.edu.
 * Web page: http://math.ucsd.edu/~sbuss/MathCG2
 */

// Use Escape or 'X' or 'x' to exit.

// These libraries are needed to link the program (Visual Studio specific)
#pragma comment(lib,"opengl32.lib")
#pragma comment(lib,"glu32.lib")
#pragma comment(lib,"glfw3.lib")
#pragma comment(lib,"glew32s.lib")
#pragma comment(lib,"glew32.lib")

// Use the static library (so glew32.dll is not needed):
#define GLEW_STATIC
#include <GL/glew.h> 
#include <GLFW/glfw3.h>

#include "GlShaderMgr.h"
#include "RgbImage.h"

bool check_for_opengl_errors();     // Function prototype (should really go in a header file)

// Enable standard input and output via printf(), etc.
// Put this include *after* the includes for glew and GLFW!
#include <stdio.h>

// ************************
// General data helping with setting up VAO (Vertex Array Objects)
//    and Vertex Buffer Objects.
// ***********************

const int NumObjects = 1;
const int iRectangle = 0;

unsigned int myVAO[NumObjects];  // Vertex Array Object - holds info about an array of vertex data;
unsigned int myVBO[NumObjects];  // Vertex Buffer Object - holds an array of data
unsigned int myEBO[NumObjects];  // Element Buffer Object - holds elements for a triangle strip

// We create one shader program: it consists of a vertex shader and a fragment shader
unsigned int shaderProgram1;
const unsigned int vertPos_loc = 0;   // Corresponds to "location = 0" in the verter shader
const unsigned int vertColor_loc = 1; // Corresponds to "location = 1" in the verter shader
const unsigned int vertTexCoords = 2;    // Corresponds to "location = 2" in the vertex shader

// *************************
// mySetupGeometries defines the scene data, especially vertex  positions and colors.
//    - It also loads all the data into the VAO (Vertex Array Object) and
//      into the VBO (Vertex Buffer Object).
// This routine is only called once to initialize the data.
// *************************
void mySetupGeometries() {

    // In this simple example, we do not use the Projection or
    //   ModelView matrices. Hence, all x, y, z positions
    //   should be in the range [-1,1].

    // Allocate Vertex Array Objects (VAOs) and Vertex Buffer Objects (VBOs).
    glGenVertexArrays(NumObjects, &myVAO[0]);
    glGenBuffers(NumObjects, &myVBO[0]);
    glGenBuffers(NumObjects, &myEBO[0]);

    // THE ONLY GEOMETRY: 
    // Specify four vertices, with a position (x,y), and texture coordinates (s,t), for each point.
    float rectVerts[] = {
        // Position (x,y)        // texture coordinates (s,t)
        -0.8f, -0.8f,            0.0f, 0.0f,                // First point
         0.8f, -0.8f,            1.0f, 0.0f,                // Second point
         0.8f,  0.8f,            1.0f, 1.0f,                // Third point
        -0.8f,  0.8f,            0.0f, 1.0f,                // Fourth point
    };

    // The elements for a triangle strip.
    unsigned int rectElements[] = { 1, 2, 0, 3 };

    // Bind (and initialize) the Vertex Array Object
    glBindVertexArray(myVAO[iRectangle]);

    // Bind and load vertex data into vertex buffer.  
    //   This may be stored in the GPU for quick access.
    glBindBuffer(GL_ARRAY_BUFFER, myVBO[iRectangle]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(rectVerts), rectVerts, GL_STATIC_DRAW);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, myEBO[iRectangle]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(rectElements), rectElements, GL_STATIC_DRAW);

    glVertexAttribPointer(vertPos_loc, 2, GL_FLOAT, GL_FALSE, 4 * sizeof(float), (void*)0);	// Info about where positions are in the VBO
    glEnableVertexAttribArray(vertPos_loc);									               // Enable the stored vertex positions
    glVertexAttribPointer(vertTexCoords, 2, GL_FLOAT, GL_FALSE, 4 * sizeof(float), (void*)(2*sizeof(float)));	// Info about where positions are in the VBO
    glEnableVertexAttribArray(vertTexCoords);									               // Enable the stored texture coordinates

	// Also optional, but perhaps a good idea so that the last VAO is not used by accident.
	glBindVertexArray(0);

    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

void myLoadTexture() {

    RgbImage theTexMap("RedLeavesTexture.bmp");

    unsigned int textureName;

    glGenTextures(1, &textureName);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, textureName);

    // Pixel alignment: each row is word aligned (aligned to a 4 byte boundary)
    //    Therefore, no need to call glPixelStore( GL_UNPACK_ALIGNMENT, ... );

    // Set wrapping parameters
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    // Set the best quality filtering options.
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

    int textureWidth = theTexMap.GetNumCols();
    int textureHeight = theTexMap.GetNumRows();
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, textureWidth, textureHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, theTexMap.ImageData());
    glGenerateMipmap(GL_TEXTURE_2D);
    
    // Make sure that the shaderProgram1 uses the GL_TEXTURE_0 texture.
    glUseProgram(shaderProgram1);
    glUniform1i(glGetUniformLocation(shaderProgram1, "theTextureMap"), 0);
}

// *************************************
// Main routine for rendering the scene
// myRenderScene() is called every time the scene needs to be redrawn.
// mySetupGeometries() has already created the vertex and buffer objects, etc.
// setup_shaders() has already created the shader programs.
// *************************************
void myRenderScene() {

	// Clear the rendering window
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// This shader program forms the texture
	glUseProgram(shaderProgram1);

	glBindVertexArray(myVAO[iRectangle]);
    glVertexAttrib3f(vertColor_loc, 1.0f, 1.0f, 1.0f);              // The color underlying the texture map is white.
    glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_INT, (void*)0);
	
    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}


void my_setup_SceneData() {
	mySetupGeometries();

    GlShaderMgr::LoadShaderSource("TextureBmpModern.glsl");
    shaderProgram1 = GlShaderMgr::CompileAndLinkAll();

    myLoadTexture();

    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

// *******************************************************
// Process all key press events.
// This routine is called each time a key is pressed or released.
// *******************************************************
void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
	if (action == GLFW_RELEASE) {
		return;			// Ignore key up (key release) events
	}
	if (key == GLFW_KEY_ESCAPE ) {
		glfwSetWindowShouldClose(window, true);
	}
}


// *************************************************
// This function is called with the graphics window is first created,
//    and again whenever it is resized.
// The Projection View Matrix is typically set here.
//    But this program does not use any transformations or matrices.
// *************************************************
void window_size_callback(GLFWwindow* window, int width, int height) {
	glViewport(0, 0, width, height);		// Draw into entire window
}

void my_setup_OpenGL() {
	
	glEnable(GL_DEPTH_TEST);	// Enable depth buffering
	glDepthFunc(GL_LEQUAL);		// Useful for multipass shaders
}

void error_callback(int error, const char* description)
{
	// Print error
	fputs(description, stderr);
}

void setup_callbacks(GLFWwindow* window) {
	// Set callback function for resizing the window
	glfwSetFramebufferSizeCallback(window, window_size_callback);

	// Set callback for key up/down/repeat events
	glfwSetKeyCallback(window, key_callback);

	// Set callbacks for mouse movement (cursor position) and mouse botton up/down events.
	// glfwSetCursorPosCallback(window, cursor_pos_callback);
	// glfwSetMouseButtonCallback(window, mouse_button_callback);
}

int main() {
	glfwSetErrorCallback(error_callback);	// Supposed to be called in event of errors. (doesn't work?)
	glfwInit();

// **********************
// Here is the main program
// **********************
#if defined(__APPLE__) || defined(__linux__)
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
#endif

	const int initWidth = 800;
	const int initHeight = 600;
	GLFWwindow* window = glfwCreateWindow(initWidth, initHeight, "TextureBmpModern", NULL, NULL);
	if (window == NULL) {
		fprintf(stdout, "Failed to create GLFW window!\n");
		glfwTerminate();
		return -1;
	}
	glfwMakeContextCurrent(window);

	glewInit();

	// Print info of GPU and supported OpenGL version
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("OpenGL version supported %s\n", glGetString(GL_VERSION));
	printf("Supported GLSL version is %s.\n", (char *)glGetString(GL_SHADING_LANGUAGE_VERSION));
    printf("Using GLEW version %s.\n", glewGetString(GLEW_VERSION));

	printf("------------------------------\n");
	printf("Press ESCAPE to exit.\n");
	
    setup_callbacks(window);
    window_size_callback(window, initWidth, initHeight);

	// Initialize OpenGL, the scene and the shaders
    my_setup_OpenGL();
	my_setup_SceneData();
 
    // Loop while program is not terminated.
	while (!glfwWindowShouldClose(window)) {
	
		myRenderScene();				// Render into the current buffer
		glfwSwapBuffers(window);		// Displays what was just rendered (using double buffering).

		// Poll events (key presses, mouse events)
		glfwWaitEvents();					// Use this if no animation.
		//glfwWaitEventsTimeout(1.0/60.0);	// Use this to animate at 60 frames/sec (timing is NOT reliable)
		// glfwPollEvents();				// Use this version when animating as fast as possible
	}

	glfwTerminate();
	return 0;
}

// If an error is found, it could have been caused by any command since the
//   previous call to check_for_opengl_errors()
// To find what generated the error, you can try adding more calls to
//   check_for_opengl_errors().
char errNames[8][36] = {
    "Unknown OpenGL error",
    "GL_INVALID_ENUM", "GL_INVALID_VALUE", "GL_INVALID_OPERATION",
    "GL_INVALID_FRAMEBUFFER_OPERATION", "GL_OUT_OF_MEMORY",
    "GL_STACK_UNDERFLOW", "GL_STACK_OVERFLOW" };
bool check_for_opengl_errors() {
    int numErrors = 0;
    GLenum err;
    while ((err = glGetError()) != GL_NO_ERROR) {
        numErrors++;
        int errNum = 0;
        switch (err) {
        case GL_INVALID_ENUM:
            errNum = 1;
            break;
        case GL_INVALID_VALUE:
            errNum = 2;
            break;
        case GL_INVALID_OPERATION:
            errNum = 3;
            break;
        case GL_INVALID_FRAMEBUFFER_OPERATION:
            errNum = 4;
            break;
        case GL_OUT_OF_MEMORY:
            errNum = 5;
            break;
        case GL_STACK_UNDERFLOW:
            errNum = 6;
            break;
        case GL_STACK_OVERFLOW:
            errNum = 7;
            break;
        }
        printf("OpenGL ERROR: %s.\n", errNames[errNum]);
    }
    return (numErrors != 0);
}